#version 430 core
struct Material {
    sampler2D diffuset;
    sampler2D speculart;
	sampler2D emissiont;
    
    vec3 diffuse;
    vec3 ambient;
    vec3 specular;
    float shininess;
};

struct DirLight {
    vec3 direction;
	
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};

struct PointLight {
    vec3 position;
    
    float constant;
    float linear;
    float quadratic;
	
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};

struct SpotLight {
    vec3 position;
    vec3 direction;
    float cutOff;
    float outerCutOff;
  
    float constant;
    float linear;
    float quadratic;
  
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;       
};

#define NR_POINT_LIGHTS 4
#define NR_SPOT_LIGHTS 2

in vec3 FragPos;
in vec3 Normal;
in vec2 TexCoords;
in vec4 FragPosLightSpace;

out vec4 color;

uniform vec3 viewPos;
uniform vec3 lightPos;
uniform Material material;
uniform DirLight dirLight;
uniform PointLight pointLights[NR_POINT_LIGHTS];
uniform SpotLight spotLights[NR_SPOT_LIGHTS] ;
uniform sampler2D texture_diffuse1;
uniform sampler2D texture_specular1;
uniform sampler2D texture_diffuse2;
uniform sampler2D texture_specular2;
uniform sampler2D shadowMap;
uniform samplerCube skybox;
uniform bool gamma;

// Function prototypes
vec3 CalcDirLight(DirLight light, vec3 normal, vec3 viewDir);
vec3 CalcPointLight(PointLight light, vec3 normal, vec3 fragPos, vec3 viewDir);
vec3 CalcSpotLight(SpotLight light, vec3 normal, vec3 fragPos, vec3 viewDir);
float ShadowCalculation(vec4 fragPosLightSpace);

void main()
{    
    // Properties
    vec3 norm = normalize(Normal);
    vec3 viewDir = normalize(viewPos - FragPos);

    vec3 result;

    // Directional Lighting
    result = CalcDirLight(dirLight, norm, viewDir);

    // Point Lights
    //for(int i = 0; i < NR_POINT_LIGHTS; i++)
        //result += CalcPointLight(pointLights[i], norm, FragPos, viewDir); 

    // Spot Light
    for(int i = 0; i < NR_SPOT_LIGHTS; i++)
        result += CalcSpotLight(spotLights[i], norm, FragPos, viewDir);   

	if (gamma) result = pow(result, vec3(1.0/2.2));
    color = vec4(result, 1.0);
}

// Calculates the color when using a directional light.
vec3 CalcDirLight(DirLight light, vec3 normal, vec3 viewDir)
{
    vec3 lightDir = normalize(-light.direction);
    
	// Diffuse shading
    float diff = max(dot(normal, lightDir), 0.0);
    
	// Specular shading
    vec3 halfwayDir = normalize(lightDir + viewDir);
    float spec = pow(max(dot(normal, halfwayDir), 0.0), material.shininess);
    
	// Combine results
    vec3 ambient = light.ambient * material.ambient;
    vec3 diffuse = light.diffuse * diff * material.diffuse;
    vec3 specular = light.specular * spec * material.specular;

	 // Textures
	vec3 diffT1 = vec3(texture(texture_diffuse1, TexCoords));
	vec3 specT1 = vec3(texture(texture_specular1, TexCoords));
	vec3 diffT2 = vec3(texture(texture_diffuse2, TexCoords));
	vec3 specT2 = vec3(texture(texture_specular2, TexCoords));

	bool dt1 = (diffT1.r > 0 || diffT1.g > 0 || diffT1.b > 0);
	bool st1 = (specT1.r > 0 || specT1.g > 0 || specT1.b > 0); 
	bool dt2 = (diffT2.r > 0 || diffT2.g > 0 || diffT2.b > 0); 
	bool st2 = (specT2.r > 0 || specT2.g > 0 || specT2.b > 0); 
	

	// Cube Map
    vec3 I = normalize(FragPos - viewPos);
    vec3 R = reflect(I, normal);
	
	// Shadows
	float shadow = ShadowCalculation(FragPosLightSpace);
	
	vec3 color = (ambient + (1.0 - shadow) * (diffuse + specular));
	if (dt1) color *= diffT1;
	if (st1) color *= specT1;
	if (dt2) color *= diffT2;
	if (st2) color *= specT2;
	color *=  texture(skybox, R).rgb;
	
	return color;
}

// Calculates the color when using a point light.
vec3 CalcPointLight(PointLight light, vec3 normal, vec3 fragPos, vec3 viewDir)
{
    vec3 lightDir = normalize(light.position - fragPos);
    
	// Diffuse shading
    float diff = max(dot(normal, lightDir), 0.0);
    
	// Specular shading
    vec3 halfwayDir = normalize(lightDir + viewDir);
    float spec = pow(max(dot(normal, halfwayDir), 0.0), material.shininess);
    
	// Attenuation
    float distance = length(light.position - fragPos);
    float attenuation = 1.0f / (light.constant + light.linear * distance + light.quadratic * (distance * distance));    
    
	// Combine results
    vec3 ambient = light.ambient * material.ambient;
    vec3 diffuse = light.diffuse * diff * material.diffuse;
    vec3 specular = light.specular * spec * material.specular;
    
	ambient *= attenuation;
    diffuse *= attenuation;
    specular *= attenuation;

	// Textures
	vec3 diffT1 = vec3(texture(texture_diffuse1, TexCoords));
	vec3 specT1 = vec3(texture(texture_specular1, TexCoords));
	vec3 diffT2 = vec3(texture(texture_diffuse2, TexCoords));
	vec3 specT2 = vec3(texture(texture_specular2, TexCoords));

	bool dt1 = (diffT1.r > 0 || diffT1.g > 0 || diffT1.b > 0);
	bool st1 = (specT1.r > 0 || specT1.g > 0 || specT1.b > 0); 
	bool dt2 = (diffT2.r > 0 || diffT2.g > 0 || diffT2.b > 0); 
	bool st2 = (specT2.r > 0 || specT2.g > 0 || specT2.b > 0); 
	
	// Cube Map
    vec3 I = normalize(FragPos - viewPos);
    vec3 R = reflect(I, normal);
	
	// Shadows
	float shadow = ShadowCalculation(FragPosLightSpace);
	
	vec3 color = (ambient + (1.0 - shadow) * (diffuse + specular));
	if (dt1) color *= diffT1;
	if (st1) color *= specT1;
	if (dt2) color *= diffT2;
	if (st2) color *= specT2;
	color *=  texture(skybox, R).rgb;

	return color;
}

// Calculates the color when using a spot light.
vec3 CalcSpotLight(SpotLight light, vec3 normal, vec3 fragPos, vec3 viewDir)
{
    vec3 lightDir = normalize(light.position - fragPos);
    
	// Diffuse shading
    float diff = max(dot(normal, lightDir), 0.0);
    
	// Specular shading
    vec3 halfwayDir = normalize(lightDir + viewDir);
    float spec = pow(max(dot(normal, halfwayDir), 0.0), material.shininess);
    
	// Attenuation
    float distance = length(light.position - fragPos);
    float attenuation = 1.0f / (light.constant + light.linear * distance + light.quadratic * (distance * distance));    
    
	// Spotlight intensity
    float theta = dot(lightDir, normalize(-light.direction)); 
    float epsilon = light.cutOff - light.outerCutOff;
    float intensity = clamp((theta - light.outerCutOff) / epsilon, 0.0, 1.0);
    
	// Combine results
    vec3 ambient = light.ambient * material.ambient;
    vec3 diffuse = light.diffuse * diff * material.diffuse;
    vec3 specular = light.specular * spec * material.specular;
    ambient *= attenuation * intensity;
    diffuse *= attenuation * intensity;
    specular *= attenuation * intensity;

	vec3 color = ambient + diffuse + specular;

	return color;
}

// Calculates Shadows
float ShadowCalculation(vec4 fragPosLightSpace) {
	vec3 projCoords = fragPosLightSpace.xyz/fragPosLightSpace.w;
	projCoords = projCoords * 0.5f + 0.5f;
	float closestDepth = texture(shadowMap, projCoords.xy).r;
	float currentDepth = projCoords.z;

	vec3 normal = normalize(Normal);
	vec3 lightDir = normalize(lightPos - FragPos);
	float bias =  max(0.05f * (1.0 - dot(normal, lightDir)), 0.005f);
	float shadow = 0.0f;
	
	// PCF
	vec2 texelSize = 1.0f / textureSize(shadowMap, 0);
	for (int x = -1; x <= 1; ++x) {
		for (int y = -1; y <= 1; ++y) {
			float pcfDepth = texture(shadowMap, projCoords.xy + vec2(x, y) * texelSize).r;
			shadow += currentDepth - bias > pcfDepth ? 1.0f : 0.0f;
		}
	}

	shadow /= 9.0f;

	if (projCoords.z > 1.0) shadow = 0.0f;

	return shadow;
}